@@ -90,34 +90,16 @@ impl ChunkQueue {
         self.shrink_to_fit();
     }
 
-    pub fn add_remote_tx(&mut self, tx: TransactionView, remote: (Cycle, PeerIndex)) {
-        if self.len() > DEFAULT_MAX_CHUNK_TRANSACTIONS {
-            return;
-        }
-
-        if self.contains_key(&tx.proposal_short_id()) {
-            return;
-        }
-
-        self.inner.insert(
-            tx.proposal_short_id(),
-            Entry {
-                tx,
-                remote: Some(remote),
-            },
-        );
-    }
-
     /// If the queue did not have this tx present, true is returned.
     ///
     /// If the queue did have this tx present, false is returned.
-    pub fn add_tx(&mut self, tx: TransactionView) -> bool {
+    pub fn add_tx(&mut self, tx: TransactionView, remote: Option<(Cycle, PeerIndex)>) -> bool {
         if self.contains_key(&tx.proposal_short_id()) {
             return false;
         }
 
         self.inner
-            .insert(tx.proposal_short_id(), Entry { tx, remote: None })
+            .insert(tx.proposal_short_id(), Entry { tx, remote })
             .is_none()
     }
 
